/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <linux/dvb/ca.h>
#include <vdr/dvbdevice.h>
#include <vdr/channels.h>

#include "common.h"
#include "cam.h"
#include "sc.h"
#include "filter.h"
#include "system.h"
#include "data.h"
#include "misc.h"

#define IDLE_SLEEP          0 // idleTime when sleeping
#define IDLE_GETCA        200 // idleTime when waiting for ca descriptors
#define IDLE_GETCA_SLOW 20000 // idleTime if no enc. system
#define IDLE_NO_SYNC      800 // idleTime when not in sync
#define IDLE_SYNC        2000 // idleTime when in sync

#define CW_REPEAT_TIME   2000 // rewrite CW after X ms
#define LOG_COUNT           3 // stop logging after X complete ECM cycles
#define CHAIN_HOLD     120000 // min. time to hold a logger chain
#define ECM_DATA_TIME    6000 // time to wait for ECM data updates
#define ECM_UPD_TIME   120000 // delay between ECM data updates
#define MAX_ECM_IDLE   300000 // delay before an idle handler can be removed
#define MAX_ECM_HOLD    15000 // delay before an idle handler stops processing

//#define DEBUG_PIDS
//#define DEBUG_AV7110
#define DEBUG_LOG
//#define DEBUG_LOG_EXTRA
#define DEBUG_LOG_STATS

// Hex-Data Logging
// LOG_DUMP to enable, rest to select output
//#define LOG_DUMP "/var/tmp/hexdump.txt"
#ifdef LOG_DUMP
//#define LOG_ECM
//#define LOG_EMM
//#define LOG_CAT
//#define LOG_PMT
#endif

#ifdef DEBUG_PIDS
#define dp(x) { x; }
#else
#define dp(x) ;
#endif

#ifdef DEBUG_LOG
#define dl(x) { (x); }
#else
#define dl(x) ;
#endif

#ifdef DEBUG_LOG_EXTRA
#define dle(x) { (x); }
#else
#define dle(x) ;
#endif

#if APIVERSNUM < 10338
#define CA_ENCRYPTED_MIN CACONFBASE
#endif

// -----------------------------------------------------------------------------

#ifdef DEBUG_AV7110
static unsigned int av7110_read(int fd, unsigned int addr)
{
  ca_pid_t arg;
  arg.pid=addr;
  ioctl(fd, CA_GET_MSG, &arg);
  return arg.index;
}

/*
static void av7110_write(int fd, unsigned int addr, unsigned int val)
{
  ca_pid_t arg;
  arg.pid=addr;
  arg.index=val;
  ioctl(fd, CA_SEND_MSG, &arg);
}
*/
#endif

#ifdef LOG_DUMP
#define PER_LINE 16
#include <stdarg.h>
void LogDump(const unsigned char *buffer, int n, const char *format, ...) __attribute__ ((format (printf,3,4)));
void LogDump(const unsigned char *buffer, int n, const char *format, ...)
{
  static cMutex logMutex;
  static FILE *f=0;

  time_t t=time(0);
  logMutex.Lock();
  if(!f) f=fopen(LOG_DUMP,"a");
  if(f) {
    struct tm tm_r;
    struct tm *tm=localtime_r(&t,&tm_r);
    char timestr[32];
    strftime(timestr,sizeof(timestr),"%b %d %T",tm);
    fprintf(f,"%s (size=%d/0x%04x) ",timestr,n,n);
    va_list ap; va_start(ap,format); vfprintf(f,format,ap); va_end(ap);
    fprintf(f,"\n");
    for(int i=0 ; i<n ;) {
      fprintf(f,"%04x:",i);
      for(int l=0 ; l<PER_LINE && i<n ; l++) fprintf(f," %02x",buffer[i++]);
      fprintf(f,"\n");
      }
    fflush(f);
    }
  logMutex.Unlock();
}
#endif

// -- cLogStats ---------------------------------------------------------------

#ifdef DEBUG_LOG_STATS

#define COUNTS  20
#define SAMPLE (30*1000)
#define AVR1   (60*1000)
#define AVR2   (4*60*1000)
#define AVR3   (10*60*1000)
#define REPORT (60*1000)

class cLogStats : public cThread {
private:
  cTimeMs sTime, repTime;
  int sCount, sIdx, sCounts[COUNTS][2];
protected:
  virtual void Action(void);
public:
  cLogStats(void);
  ~cLogStats();
  void Count(void);
  };

static cMutex logstatsMutex;
static cLogStats *logstats=0;

void LogStatsUp(void)
{
  logstatsMutex.Lock();
  if(!logstats) logstats=new cLogStats;
  logstatsMutex.Unlock();
}

void LogStatsDown(void)
{
  logstatsMutex.Lock();
  if(logstats) { delete logstats; logstats=0; }
  logstatsMutex.Unlock();
}

cLogStats::cLogStats(void)
{
  sCount=sIdx=0;
  for(int i=0; i<COUNTS; i++) { sCounts[i][0]=0; sCounts[i][1]=SAMPLE; }
  SetDescription("logger stats");
  Start();
}

cLogStats::~cLogStats()
{
  Cancel(2);
}

void cLogStats::Count(void)
{
  sCount++;
}

void cLogStats::Action(void)
{
  while(Running()) {
    cCondWait::SleepMs(50);
    if(sTime.Elapsed()>SAMPLE) {
      sCounts[sIdx][0]=sCount;          sCount=0;
      sCounts[sIdx][1]=sTime.Elapsed(); sTime.Set();
      if(++sIdx >= COUNTS) sIdx=0;
      }
    if(repTime.Elapsed()>REPORT) {
      repTime.Set();
      if(sCounts[(sIdx+COUNTS-1)%COUNTS][0]>0) {
        printf("logstats: EMM packet load average (%d/%d/%dmin)",AVR1/60000,AVR2/60000,AVR3/60000);
        int s=0, t=0;
        for(int i=1; i<=COUNTS; i++) {
          s+=sCounts[(sIdx+COUNTS-i)%COUNTS][0];
          t+=sCounts[(sIdx+COUNTS-i)%COUNTS][1];
          if(i==(AVR1/SAMPLE) || i==(AVR2/SAMPLE) || i==(AVR3/SAMPLE))
            printf(" %4d",(int)((float)s/(float)t*1000.0));
          }
        printf(" pks/s\n");
        }
      }
    }
}

#endif

// -- cLogChain ----------------------------------------------------------------

class cLogChain : public cSimpleItem {
public:
  int cardNum, caid;
  bool softCSA, active, delayed;
  cTimeMs delay;
  cPids pids;
  cSimpleList<cSystem> systems;
  //
  cLogChain(int CardNum, bool soft);
  void Process(int pid, unsigned char *data);
  bool Parse(const unsigned char *cat);
  };

cLogChain::cLogChain(int CardNum, bool soft)
{
  cardNum=CardNum; softCSA=soft; active=delayed=false;
}

void cLogChain::Process(int pid, unsigned char *data)
{
  if(active) {
    for(cSystem *sys=systems.First(); sys; sys=systems.Next(sys))
      sys->ProcessEMM(pid,caid,data);
    }
}

bool cLogChain::Parse(const unsigned char *cat)
{
  if(cat[0]==0x09) {
    caid=WORD(cat,2,0xFFFF);
    dl(printf("logger %d: chain caid %04x",cardNum,caid))
    cSystem *sys;
    if(systems.Count()>0) {
      dl(printf(" ++"))
      for(sys=systems.First(); sys; sys=systems.Next(sys))
        sys->ParseCAT(&pids,cat);
      }
    else {
      dl(printf(" ->"))
      int Pri=0;
      while((sys=cSystems::FindBySysId(caid,!softCSA,Pri))) {
        Pri=sys->Pri();
        if(sys->HasLogger()) {
          sys->CardNum(cardNum);
          sys->ParseCAT(&pids,cat);
          systems.Add(sys);
          dl(printf(" %s(%d)",sys->Name(),sys->Pri()))
          }
        else
          delete sys;
        }
      }
#ifdef DEBUG_LOG
    if(systems.Count()==0) printf(" none available");
    for(cPid *pid=pids.First(); pid; pid=pids.Next(pid))
       printf(" [%04x-%02x/%02x/%02x]",pid->pid,pid->sct,pid->mask,pid->mode);
    printf("\n");
#endif
    if(systems.Count()>0 && pids.Count()>0)
      return true;
    }
  return false;
}

// -- cLogger ------------------------------------------------------------------

class cLogger : public cAction {
private:
  int cardNum;
  bool softCSA, up;
  cSimpleList<cLogChain> chains;
  cSimpleList<cEcmInfo> active;
  //
  cPidFilter *catfilt;
  int catVers;
  //
  cPidFilter *AddFilter(int Pid, int Section, int Mask, int Mode, int IdleTime, bool Crc);
  void SetChains(void);
  void ClearChains(void);
  void StartChain(cLogChain *chain);
  void StopChain(cLogChain *chain, bool force);
protected:
  virtual void Process(cPidFilter *filter, unsigned char *data, int len);
public:
  cLogger(int CardNum, bool soft);
  virtual ~cLogger();
  void EcmStatus(const cEcmInfo *ecm, bool on);
  void Up(void);
  void Down(void);
  };

cLogger::cLogger(int CardNum, bool soft)
:cAction("logger",CardNum)
{
  cardNum=CardNum; softCSA=soft;
  catfilt=0; up=false;
  Priority(10);
}

cLogger::~cLogger()
{
  Down();
}

void cLogger::Up(void)
{
  Lock();
  if(!up) {
    dle(printf("logger %d: UP\n",cardNum))
    catVers=-1;
    catfilt=AddFilter(1,0x01,0xFF,0,0,true);
    up=true;
    }
  Unlock();
}

void cLogger::Down(void)
{
  Lock();
  if(up) {
    dle(printf("logger %d: DOWN\n",cardNum))
    ClearChains();
    DelAllFilter();
    catfilt=0; up=false;
    }
  Unlock();
}

void cLogger::EcmStatus(const cEcmInfo *ecm, bool on)
{
  Lock();
  dle(printf("logger %d: ecm prgid=%d caid=%04x prov=%.4x %s\n",cardNum,ecm->prgId,ecm->caId,ecm->provId,on ? "active":"inactive"))
  cEcmInfo *e;
  if(on) {
    e=new cEcmInfo(ecm);
    active.Add(e);
    if(!up) Up();
    }
  else {
    for(e=active.First(); e; e=active.Next(e))
      if(e->Compare(ecm)) {
        active.Del(e);
        break;
        }
    }
  SetChains();
  Unlock();
}

void cLogger::SetChains(void)
{
  for(cLogChain *chain=chains.First(); chain; chain=chains.Next(chain)) {
    bool act=false;
    if(ScSetup.AutoUpdate>1) act=true;
    else if(ScSetup.AutoUpdate==1) {
      for(cEcmInfo *e=active.First(); e; e=active.Next(e))
        if((e->emmCaId && chain->caid==e->emmCaId) || chain->caid==e->caId) {
          act=true; break;
          }
       }
    if(act) StartChain(chain);
    else StopChain(chain,false);
    }
}

void cLogger::ClearChains(void)
{
  for(cLogChain *chain=chains.First(); chain; chain=chains.Next(chain))
    StopChain(chain,true);
  chains.Clear();
}

void cLogger::StartChain(cLogChain *chain)
{
  if(chain->delayed)
    dle(printf("logger %d: restarting delayed chain %04x\n",cardNum,chain->caid))
  chain->delayed=false;
  if(!chain->active) {
    dl(printf("logger %d: starting chain %04x\n",cardNum,chain->caid))
    chain->active=true;
    for(cPid *pid=chain->pids.First(); pid; pid=chain->pids.Next(pid)) {
      cPidFilter *filter=AddFilter(pid->pid,pid->sct,pid->mask,pid->mode,CHAIN_HOLD/8,false);
      if(filter) {
        filter->userData=(void *)chain;
        pid->filter=filter;
        }
      }
    }
}

void cLogger::StopChain(cLogChain *chain, bool force)
{
  if(chain->active) {
    if(force || (chain->delayed && chain->delay.TimedOut())) {
      dl(printf("logger %d: stopping chain %04x\n",cardNum,chain->caid))
      chain->active=false;
      for(cPid *pid=chain->pids.First(); pid; pid=chain->pids.Next(pid)) {
        cPidFilter *filter=pid->filter;
        if(filter) {
          DelFilter(filter);
          pid->filter=0;
          }
        }
      }
    else if(!chain->delayed) {
      dle(printf("logger %d: delaying chain %04x\n",cardNum,chain->caid))
      chain->delayed=true;
      chain->delay.Set(CHAIN_HOLD);
      }
    }
}

cPidFilter *cLogger::AddFilter(int Pid, int Section, int Mask, int Mode, int IdleTime, bool Crc)
{
  cPidFilter *filter=NewFilter(IdleTime);
  if(filter) {
    if(Pid>1) filter->SetBuffSize(32768);
    filter->Start(Pid,Section,Mask,Mode,Crc);
    dle(printf("logger %d: added filter pid=0x%.4x sct=0x%.2x/0x%.2x/0x%.2x idle=%d crc=%d\n",cardNum,Pid,Section,Mask,Mode,IdleTime,Crc))
    }
  else d(printf("logger %d: no free slot or filter failed to open\n",cardNum))
  return filter;
}

void cLogger::Process(cPidFilter *filter, unsigned char *data, int len)
{
  if(data && len>0) {
    if(filter==catfilt) {
      int vers=(data[5]&0x3E)>>1;
      if(data[0]==0x01 && vers!=catVers) {
        dle(printf("logger %d: got CAT version %02x\n",cardNum,vers))
        catVers=vers;
#ifdef LOG_CAT
        LogDump(data,len,"CAT vers %02x",catVers);
#endif
        ClearChains();
        for(int i=8; i<len-4; i+=data[i+1]+2) {
          if(data[i]==0x09) {
            int caid=WORD(data,i+2,0xFFFF);
            cLogChain *chain;
            for(chain=chains.First(); chain; chain=chains.Next(chain))
              if(chain->caid==caid) break;
            if(chain)
              chain->Parse(&data[i]);
            else {
              chain=new cLogChain(cardNum,softCSA);
              if(chain->Parse(&data[i]))
                chains.Add(chain);
              else
                delete chain;
              }
            }
          }
        SetChains();
        }
      }
    else {
#ifdef LOG_EMM
      LogDump(data,len,"EMM pid 0x%04x",filter->Pid());
#endif
#ifdef DEBUG_LOG_STATS
      if(logstats) logstats->Count();
#endif
      if(SCT_LEN(data)==len) {
        cLogChain *chain=(cLogChain *)(filter->userData);
        if(chain) {
          chain->Process(filter->Pid(),data);
          if(chain->delayed) StopChain(chain,false);
          }
        }
      else d(printf("logger %d: incomplete section %d != %d\n",cardNum,len,SCT_LEN(data)))
      }
    }
  else {
    cLogChain *chain=(cLogChain *)(filter->userData);
    if(chain && chain->delayed) StopChain(chain,false);
    }
}

// -- cSids --------------------------------------------------------------------

#define MAX_SIDS 16

class cSids {
private:
  int num, sids[MAX_SIDS];
public:
  cSids(void);
  void AddSid(int sid);
  void Copy(const cSids *s);
  void Clear(void) { num=0; }
  bool HasSid(int sid) const;
  int Count(void) const { return num; }
  int GetSid(int idx) const;
  bool Compare(const cSids *s, bool common) const;
#if defined(DEBUG) | defined(DEBUG_PIDS)
  cString Print(void) const;
#endif
  };

cSids::cSids(void)
{
  Clear();
}

void cSids::AddSid(int sid)
{
  if(num<MAX_SIDS) {
    int i;
    for(i=0; i<num; i++) {
      if(sids[i]==sid) return;
      if(sids[i]>sid) break;
      }
    if(i<num)
      memmove(&sids[i+1],&sids[i],(sizeof(sids)-sizeof(int))-sizeof(int)*i);
    sids[i]=sid;
    num++;
    }
}

void cSids::Copy(const cSids *s)
{
  num=s->num;
  memcpy(sids,s->sids,sizeof(sids));
}

bool cSids::HasSid(int sid) const
{
  for(int i=0; i<num; i++)
    if(sids[i]==sid) return true;
  return false;
}

int cSids::GetSid(int idx) const
{
  return (idx<num) ? sids[idx] : -1;
}

bool cSids::Compare(const cSids *s, bool common) const
{
  if(!common)
    return num==s->num && memcmp(sids,s->sids,sizeof(int)*num)==0;

  for(int i=0; i<num; i++) {
    for(int j=0; j<s->num; j++)
      if(sids[i]==s->sids[j]) return true;
    }
  return false;
}

#if defined(DEBUG) | defined(DEBUG_PIDS)
cString cSids::Print(void) const
{
  if(num>0) {
    char buf[512];
    int q=0;
    for(int i=0; i<num; i++)
      q+=snprintf(&buf[q],sizeof(buf)-q,i?",%d":"%d",sids[i]);
    return buf;
    }
  else return "<none>";
}
#endif

// -- cEcmData -----------------------------------------------------------------

class cEcmData : public cEcmInfo {
private:
  bool del;
public:
  cEcmData(void);
  cEcmData(cEcmInfo *e);
  bool Save(FILE *f);
  bool Parse(const char *buf);
  void Delete(void) { del=true; }
  bool IsDeleted(void) const { return del; }
  };

cEcmData::cEcmData(void)
:cEcmInfo()
{
  del=false;
}

cEcmData::cEcmData(cEcmInfo *e)
:cEcmInfo(e)
{
  del=false;
}

bool cEcmData::Parse(const char *buf)
{
  char Name[64];
  int nu=0, num;
  Name[0]=0;
  if(sscanf(buf,"%d:%x:%x:%63[^:]:%x/%x:%x:%x/%x:%d%n",&prgId,&source,&transponder,Name,&caId,&emmCaId,&provId,&ecm_pid,&ecm_table,&nu,&num)>=9) {
    SetName(Name);
    const char *line=buf+num;
    if(nu>0 && *line++==':') {
      unsigned char dat[nu];
      if(GetHex(line,dat,nu,true)==nu) AddData(dat,nu);
      }
    return true;
    }
  return false;
}

bool cEcmData::Save(FILE *f)
{
  fprintf(f,"%d:%x:%x:%s:%x/%x:%x:%x/%x",prgId,source,transponder,name,caId,emmCaId,provId,ecm_pid,ecm_table);
  if(data) {
    char str[dataLen*2+2];
    fprintf(f,":%d:%s\n",dataLen,HexStr(str,data,dataLen));
    }
  else
    fprintf(f,":0\n");
  return ferror(f)==0;
}

// -- cEcmCache ----------------------------------------------------------------

cEcmCache ecmcache;

cEcmCache::cEcmCache(void)
:cLoader("ECM")
{}

void cEcmCache::New(cEcmInfo *e)
{
  Lock();
  cEcmData *dat;
  if(!(dat=Exists(e))) {
    dat=new cEcmData(e);
    Add(dat);
    Modified();
    d(printf("ecmcache: added prgId=%d source=%x transponder=%x ecm=%x/%x\n",e->prgId,e->source,e->transponder,e->ecm_pid,e->ecm_table))
    }
  else {
    if(strcasecmp(e->name,dat->name)) {
      dat->SetName(e->name);
      Modified();
      }
    if(dat->Update(e))
      Modified();
    }
  e->SetCached();
  Unlock();
}

cEcmData *cEcmCache::Exists(cEcmInfo *e)
{
  for(cEcmData *dat=First(); dat; dat=Next(dat))
    if(!dat->IsDeleted() && dat->Compare(e)) return dat;
  return 0;
}

int cEcmCache::GetCached(cSimpleList<cEcmInfo> *list, const cSids *sids, int Source, int Transponder)
{
  int n=0;
  list->Clear();
  Lock();
  cEcmData *dat=First();
  while(dat) {
    if(!dat->IsDeleted() && sids->HasSid(dat->prgId) && dat->source==Source && dat->transponder==Transponder) {
      cEcmInfo *e=new cEcmInfo(dat);
      if(e) {
        d(printf("ecmcache: from cache: system %s (%04x) id %04x with ecm %x/%x\n",e->name,e->caId,e->provId,e->ecm_pid,e->ecm_table))
        e->SetCached();
        list->Add(e);
        n++;
        }
      }
    dat=Next(dat);
    }
  Unlock();
  return n;
}

void cEcmCache::Delete(cEcmInfo *e)
{
  Lock();
  cEcmData *dat=Exists(e);
  if(dat) {
    dat->Delete();
    Modified();
    d(printf("ecmcache: invalidated prgId=%d source=%x transponder=%x ecm=%x/%x\n",dat->prgId,dat->source,dat->transponder,dat->ecm_pid,dat->ecm_table))
    }
  Unlock();
}

void cEcmCache::Load(void)
{
  Clear();
}

bool cEcmCache::Save(FILE *f)
{
  bool res=true;
  Lock();
  cEcmData *dat=First();
  while(dat) {
    if(!dat->IsDeleted() && !dat->Save(f)) { res=false; break; }
    dat=Next(dat);
    }
  Modified(!res);
  Unlock();
  return res;
}

bool cEcmCache::ParseLine(const char *line, bool fromCache)
{
  bool res=false;
  cEcmData *dat=new cEcmData;
  if(dat && dat->Parse(line)) {
    if(!Exists(dat)) { Add(dat); dat=0; }
    res=true;
    }
  delete dat;
  return res;
}

// -- cEcmSys ------------------------------------------------------------------

class cEcmPri : public cSimpleItem {
public:
  cEcmInfo *ecm;
  int pri, sysIdent;
  };

// -- cEcmHandler --------------------------------------------------------------

class cEcmHandler : public cSimpleItem, public cAction {
private:
  int cardNum, cwIndex;
  cSids sids;
  cCam *cam;
  char *id;
  //
  int numPids;
  unsigned short pids[MAX_HLR_PIDS];
  cTimeMs idleTime;
  //
  cSystem *sys;
  cPidFilter *filter;
  int filterCwIndex, filterSource, filterTransponder;
  cSids filterSids;
  unsigned char lastCw[16];
  bool sync, noKey, trigger;
  cMutex triggerMutex;
  int triggerMode;
  int mode, count;
  cTimeMs lastsync, startecm, resendTime;
  unsigned int cryptPeriod;
  unsigned char parity;
  cMsgCache failed;
  //
  cSimpleList<cEcmInfo> ecmList;
  cSimpleList<cEcmPri> ecmPriList;
  cEcmInfo *ecm;
  cEcmPri *ecmPri;
  cTimeMs ecmUpdTime;
  //
  int dolog;
  //
  void DeleteSys(void);
  void NoSync(bool clearParity);
  cEcmInfo *NewEcm(void);
  cEcmInfo *JumpEcm(void);
  void StopEcm(void);
  bool UpdateEcm(void);
  void EcmOk(void);
  void EcmFail(void);
  void ParseCAInfo(int sys);
  void AddEcmPri(cEcmInfo *n);
protected:
  virtual void Process(cPidFilter *filter, unsigned char *data, int len);
public:
  cEcmHandler(cCam *Cam, int CardNum, cSids *Sids, int cwindex);
  virtual ~cEcmHandler();
  void AddPid(int pid);
  bool DelPid(int pid);
  int HasPid(int pid);
  bool IsIdle(void) const { return numPids==0; }
  bool IsRemoveable(void);
  void SetSids(cSids *Sids);
  const cSids *Sids(void) const { return &sids; }
  void ShiftCwIndex(int cwindex);
  int CwIndex(void) const { return cwIndex; }
  char *CurrentKeyStr(void) const;
  const char *Id(void) const { return id; }
  };

cEcmHandler::cEcmHandler(cCam *Cam, int CardNum, cSids *Sids, int cwindex)
:cAction("ecmhandler",CardNum)
,failed(32,0)
{
  cam=Cam;
  cardNum=CardNum;
  cwIndex=cwindex;
  sys=0; filter=0; ecm=0; ecmPri=0; numPids=0; mode=-1;
  trigger=false; triggerMode=-1;
  filterSource=filterTransponder=0; filterCwIndex=-1;
  asprintf(&id,"ecmhandler %d.%d",cardNum,cwindex);
  SetSids(Sids);
}

cEcmHandler::~cEcmHandler()
{
  Lock();
  StopEcm();
  DelAllFilter(); // delete filters before sys for multi-threading reasons
  DeleteSys();
  Unlock();
  free(id);
}

void cEcmHandler::SetSids(cSids *Sids)
{
  if(numPids>0) {
    d(printf("%s: is not idle on SetSid(). Active pids are:",id))
    for(int i=0; i<numPids; i++) d(printf(" %04x",pids[i]));
    d(printf("\n"))
    }
  numPids=0;
  idleTime.Set();

  if(!sids.Compare(Sids,false)) {
    if(Sids->Count()>0) d(printf("%s: setting new SID %s\n",id,*Sids->Print()))
    triggerMutex.Lock();
    if(trigger && filterSids.Compare(Sids,true)) filterSids.Clear();
    trigger=true;
    sids.Copy(Sids);
    triggerMutex.Unlock();
    if(filter) filter->Wakeup();
    }
  else if(sids.Count()>0) d(printf("%s: SID unchanged\n",id))
}

void cEcmHandler::ShiftCwIndex(int cwindex)
{
  if(cwIndex!=cwindex) {
    dp(printf("%s: shifting cwIndex from %d to %d\n",id,cwIndex,cwindex))
    free(id);
    asprintf(&id,"ecmhandler %d.%d",cardNum,cwindex);
    triggerMutex.Lock();
    trigger=true;
    cwIndex=cwindex;
    triggerMutex.Unlock();
    if(filter) filter->Wakeup();
    }
}

bool cEcmHandler::IsRemoveable(void)
{
  return IsIdle() && idleTime.Elapsed()>MAX_ECM_IDLE;
}

void cEcmHandler::AddPid(int pid)
{
  if(numPids<MAX_HLR_PIDS) {
    dp(printf("%s: attaching pid %04x for idx %d\n",id,pid,cwIndex))
    pids[numPids]=pid; numPids++;
    cam->SetCWIndex(pid,cwIndex);

    if(numPids==1) {
      d(printf("%s: is no longer idle\n",id))
      triggerMutex.Lock();
      trigger=true;
      triggerMode=0;
      triggerMutex.Unlock();
      if(!filter) {
        filter=NewFilter(IDLE_SLEEP);
        if(!filter) d(printf("%s: failed to open ECM filter\n",id))
        }
      if(filter) filter->Wakeup();
      if(cam->IsSoftCSA()) cSoftCAM::AddConcurrent();
      }
    }
  else d(printf("%s: pid table overflow\n",id))
  return;
}

int cEcmHandler::HasPid(int pid)
{
  for(int idx=numPids-1; idx>=0; idx--)
    if(pids[idx]==pid) return idx;
  return -1;
}

bool cEcmHandler::DelPid(int pid)
{
  int idx=HasPid(pid);
  if(idx>=0) {
    dp(printf("%s: detaching pid %04x\n",id,pid))
    pids[idx]=pids[numPids-1]; numPids--;
    if(IsIdle() && filter) {
      d(printf("%s: is idle now\n",id))
      idleTime.Set();
      if(cam->IsSoftCSA()) cSoftCAM::DelConcurrent();
      }
    return true;
    }
  return false;
}

void cEcmHandler::Process(cPidFilter *filter, unsigned char *data, int len)
{
  triggerMutex.Lock();
  if(trigger) {
    d(printf("%s: triggered SID %s/%s idx %d/%d mode %d/%d %s\n",
      id,*filterSids.Print(),*sids.Print(),filterCwIndex,cwIndex,mode,triggerMode,(mode==3 && sync)?"sync":"-"))
    trigger=false;
    if(!filterSids.Compare(&sids,false)) {
      filterSids.Copy(&sids);
      filterSource=cam->Source();
      filterTransponder=cam->Transponder();
      filterCwIndex=cwIndex;
      noKey=true; mode=0;
      }
    else {
      if(filterCwIndex!=cwIndex) {
        filterCwIndex=cwIndex;
        triggerMutex.Unlock();
        cam->Lock();
        if(mode==3 && sync)
          cam->WriteCW(filterCwIndex,lastCw,true);
        for(int idx=numPids-1; idx>=0; idx--)
          cam->SetCWIndex(pids[idx],filterCwIndex);
        cam->Unlock();
        triggerMutex.Lock();
        }
      if(mode<triggerMode) mode=triggerMode;
      }
    triggerMode=-1;
    }
  triggerMutex.Unlock();

  switch(mode) {
    case -1:
      filter->SetIdleTime(IDLE_SLEEP);
      break;

    case 0:
      StopEcm();
      if(numPids==0) { mode=-1; break; }

      dolog=LOG_COUNT;
      NewEcm();
      filter->SetIdleTime(IDLE_GETCA);
      startecm.Set();
      mode=1;
      // fall through

    case 1:
      if(!ecm && !JumpEcm()) {
        if(startecm.Elapsed()>IDLE_GETCA_SLOW) {
          if(IsIdle()) { mode=0; break; }
          d(printf("%s: no encryption system found\n",id))
          filter->SetIdleTime(IDLE_GETCA_SLOW/4);
          startecm.Set();
          }
        break;
        }
      mode=4;
      // fall through

    case 4:
    case 5:
      NoSync(mode==4);
      failed.Clear();
      filter->SetIdleTime(IDLE_NO_SYNC/2);
      lastsync.Set();
      cryptPeriod=60*1000;
      mode=2;
      // fall through
          
    case 2:
      if(sys->NeedsData()) {
        if(!UpdateEcm()) {
          if(lastsync.Elapsed()<ECM_DATA_TIME) break;
          d(printf("%s: no ecm extra data update (waited %lld ms)\n",id,lastsync.Elapsed()))
          }
        if(lastsync.Elapsed()>IDLE_NO_SYNC/4 && dolog)
          d(printf("%s: ecm extra data update took %lld ms\n",id,lastsync.Elapsed()))
        }
      filter->SetIdleTime(IDLE_NO_SYNC);
      mode=3;
      // fall through

    case 3:
      {
      bool resend=false, cwok=false;
      if(resendTime.TimedOut()) {
        resend=sync; resendTime.Set(8*24*60*60*1000);
        }

#ifdef DEBUG_AV7110
      if(startecm.Elapsed()<3*60*1000) cam->Dump();
#endif

      if(data && len>0) {
#ifdef LOG_ECM
        LogDump(data,len,"ECM sys 0x%04x id 0x%02x pid 0x%04x",ecm->caId,ecm->provId,filter->Pid());
#endif

        if(SCT_LEN(data)==len) {
          depDUMP(data,16,"%s: ECM",id);
          int n;
          if(!(n=sys->CheckECM(ecm,data,sync))) {
            if(resend || parity!=(data[0]&1)) {
              int ecmid;
              cTimeMs procTime;
              cwok=(ecmid=failed.Get(data,len,0))>=0 && sys->ProcessECM(ecm,data);
              n=(ecmid>0)?failed.Cache(ecmid,cwok,0):99;
              sys->CheckECMResult(ecm,data,cwok);
              if(cwok) {
                parity=data[0]&1;
                }
              else {
                if(procTime.Elapsed()>6000) {
                  d(printf("%s: filter flush (elapsed %d)\n",id,(int)procTime.Elapsed()))
                  filter->Flush();
                  }
                if(n>=2) { count++; if(n==2) count++; }
                parity=0xFF;
                if(sync && lastsync.Elapsed()>cryptPeriod*2) {
                  d(printf("%s: lost sync (period %d, elapsed %d)\n",id,cryptPeriod,(int)lastsync.Elapsed()))
                  NoSync(true);
                  }
                }
              dep(printf("%s: ECM proc (%s) cwok=%d ecmid=%d n=%d sync=%d parity=%d count=%d ELA=%d\n",
                    id,sys->Name(),cwok,ecmid,n,sync,parity,count,(int)procTime.Elapsed()))
              }
            }
          else {
            dep(printf("%s: ECM check result %d\n",id,n))
            switch(n) {
              case 1: NoSync(true); break;
              case 2: count++; break;
              case 3: break;
              }
            }
          }
        else {
          d(printf("%s: incomplete section %d != %d\n",id,len,SCT_LEN(data)))
          count++;
          }
        }
      else if(sys->Constant()) {
        if(sys->ProcessECM(ecm,NULL)) {
          cwok=true;
          if(sync) filter->SetIdleTime(IDLE_SYNC*10); 
          }
        else count++;
        }
      else count++;

      if(cwok) {
        dolog=LOG_COUNT; sys->DoLog(true);
        cam->WriteCW(filterCwIndex,sys->CW(),resend || !sync);
        memcpy(lastCw,sys->CW(),sizeof(lastCw));
        noKey=false; count=0;
        UpdateEcm(); EcmOk();
        if(!sync) {
          sync=true;
          filter->SetIdleTime(IDLE_SYNC);
          d(printf("%s: correct key found\n",id))
          if(!cam->IsSoftCSA())
            resendTime.Set(CW_REPEAT_TIME);
          }
        else if(!resend)
          cryptPeriod=max(5000,min(60000,(int)lastsync.Elapsed()));
        lastsync.Set();
        }

      if(!sync && !trigger) {
        if(count>=sys->MaxEcmTry()) {
          EcmFail(); JumpEcm();
          mode=4;
          if(!ecm) {
            JumpEcm();
            if(!ecm) { // this should not happen!
              d(printf("%s: BUG: empty ecm list in sync loop\n",id))
              mode=0; break;
              }
            // if we looped through all systems, we wait until the next parity
            // change before we try again.
            if(dolog!=LOG_COUNT && data) { parity=data[0]&1; mode=5; }
            if(dolog && !--dolog) {
#ifdef DEBUG
              sys->DoLog(false);
              printf("%s: stopping console log until valid key is found\n",id);
#endif
              }
            }
          break;
          }
        }
        
      if(IsIdle() && idleTime.Elapsed()>MAX_ECM_HOLD) {
        d(printf("%s: hold timeout expired\n",id))
        mode=0;
        }

      break;
      }
    }
}

void cEcmHandler::NoSync(bool clearParity)
{
  if(clearParity) parity=0xFF;
  count=0; sync=false;
}

void cEcmHandler::DeleteSys(void)
{
  delete sys; sys=0;
}

char *cEcmHandler::CurrentKeyStr(void) const
{
  if(noKey || !sys) return 0;
  return strdup(sys->CurrentKeyStr());
}

cEcmInfo *cEcmHandler::NewEcm(void)
{
  ecmcache.GetCached(&ecmList,&filterSids,filterSource,filterTransponder);
  ecmPriList.Clear();
  for(cEcmInfo *n=ecmList.First(); n; n=ecmList.Next(n)) AddEcmPri(n);
  ecm=0; ecmPri=0;
  return JumpEcm();
}

void cEcmHandler::AddEcmPri(cEcmInfo *n)
{
  int ident, pri=0;
  while((ident=cSystems::FindIdentBySysId(n->caId,!cam->IsSoftCSA(),pri))>0) {
    cEcmPri *ep=new cEcmPri;
    if(ep) {
      ep->ecm=n; 
      ep->pri=pri;
      ep->sysIdent=ident;
      if(n->Cached()) ep->pri+=20;

      // keep ecmPriList sorted
      cEcmPri *eppp, *epp=ecmPriList.First();
      if(!epp || epp->pri<ep->pri)
        ecmPriList.Ins(ep);
      else {
        do {
          eppp=ecmPriList.Next(epp);
          if(!eppp || eppp->pri<ep->pri) {
            ecmPriList.Add(ep,epp);
            break;
            }
          } while((epp=eppp));
        }
      }
    }
}

void cEcmHandler::StopEcm(void)
{
  filter->Stop(); filter->Flush();
  if(ecm) cam->LogEcmStatus(ecm,false);
  DeleteSys();
}

bool cEcmHandler::UpdateEcm(void)
{
  if(!ecm->Data() || ecmUpdTime.TimedOut()) {
#ifdef DEBUG
    bool log=dolog;
    dolog=(sys && sys->NeedsData() && ecm->Data()==0);
    if(dolog) printf("%s: try to update ecm extra data\n",id);
#endif
    ParseCAInfo(ecm->caId);
    ecmUpdTime.Set(ECM_UPD_TIME);
#ifdef DEBUG
    dolog=log;
#endif
    if(!ecm->Data()) return false;
    }
  return true;
}

cEcmInfo *cEcmHandler::JumpEcm(void)
{
  noKey=true;
  if(!ecmPri) {
    ParseCAInfo(0xFFFF); // all systems
    ecmPri=ecmPriList.First();
    }
  else ecmPri=ecmPriList.Next(ecmPri);
  if(ecmPri) {
    if(ecmPri->ecm!=ecm) {
      StopEcm(); ecmUpdTime.Set();
      ecm=ecmPri->ecm;
      filter->Start(ecm->ecm_pid,ecm->ecm_table,0xfe,0,false);
      cam->LogEcmStatus(ecm,true);
      }
    else {
      DeleteSys();
      filter->Flush();
      }
    sys=cSystems::FindBySysIdent(ecmPri->sysIdent);
    if(!sys) {
      if(dolog) d(printf("%s: internal error: no system found for ident %04x (caid %04x pri %d)\n",id,ecmPri->sysIdent,ecmPri->ecm->caId,ecmPri->pri))
      return JumpEcm();
      }
    sys->DoLog(dolog!=0); sys->CardNum(cardNum);
    failed.SetMaxFail(sys->MaxEcmTry());

    if(dolog) d(printf("%s: try system %s (%04x) id %04x with ecm %x%s (pri=%d)\n",
                     id,sys->Name(),ecm->caId,ecm->provId,ecm->ecm_pid,ecm->Cached()?" (cached)":"",sys->Pri()))
    }
  else {
    StopEcm();
    ecm=0;
    }
  return ecm;
}

void cEcmHandler::EcmOk(void)
{
  ecm->SetName(sys->Name());
  ecm->Fail(false);
  ecmcache.New(ecm);
  cEcmInfo *e=ecmList.First();
  while(e) {
    if(e->Cached() && e->Failed()) ecmcache.Delete(e);
    e=ecmList.Next(e);
    }
  cSoftCAM::Save();
}

void cEcmHandler::EcmFail(void)
{
  ecm->Fail(true);
}

void cEcmHandler::ParseCAInfo(int sys)
{
  unsigned char buff[2048];
  unsigned short casys[2];
  casys[0]=sys;
  casys[1]=0;
  for(int idx=0, sid; (sid=filterSids.GetSid(idx))>=0; idx++) {
    bool streamFlag;
    int len=GetCaDescriptors(filterSource,filterTransponder,sid,casys,sizeof(buff),buff,streamFlag);
    if(len>0) {
      if(dolog) d(printf("%s: got CaDescriptors for SID %04X (len=%d)\n",id,sid,len))
#ifdef LOG_PMT
      LogDump(buff,len,"PMT");
#endif
      int index=0;
      while(index<len) {
        int s=buff[index+1];
        if(buff[index]==0x09) {
#ifdef DEBUG
          if(dolog) {
            printf("%s: descriptor",id);
            for(int i=0 ; i<s ; i++) printf(" %02x",buff[index+2+i]);
            printf("\n");
            }
#endif
          int sysId=WORD(buff,index+2,0xFFFF);
          int sysPri=0;
          cSystem *sys;
          while((sys=cSystems::FindBySysId(sysId,!cam->IsSoftCSA(),sysPri))) {
            sysPri=sys->Pri();
            cSimpleList<cEcmInfo> ecms;
            sys->ParseCADescriptor(&ecms,sysId,&buff[index+2],s);
            delete sys;
            if(ecms.Count()) {
              cEcmInfo *n;
              while((n=ecms.First())) {
                ecms.Del(n,false);
                if(dolog) d(printf("%s: found %04x (%s) id %04x with ecm %x ",id,n->caId,n->name,n->provId,n->ecm_pid))
                cEcmInfo *e=ecmList.First();
                while(e) {
                  if(e->ecm_pid==n->ecm_pid) {
                    if(e->caId==n->caId && e->provId==n->provId) {
                      if(n->Data()) {
                        if(e->Update(n) && dolog) d(printf("(updated) "))
                        }
                      if(dolog) d(printf("(already present)\n"))
                      delete n; n=0;
                      break;
                      }
                    else {
                      e->Fail(true);
                      if(dolog) d(printf("(dup) "))
                      }
                    }
                  e=ecmList.Next(e);
                  }
                if(n) {
                  if(dolog) d(printf("(new)\n"))
                  n->SetSource(sid,filterSource,filterTransponder);
                  ecmList.Add(n);
                  AddEcmPri(n);
                  }
                }
              break;
              }
            }
          if(sysPri==0 && dolog) d(printf("%s: no module available for system %04x\n",id,sysId))
          }
        s+=2; index+=s;
        }
      }
    else if(len<0)
      d(printf("%s: CA parse buffer overflow\n",id))
    }
#ifdef DEBUG_ECM_PROG
  if(sys==0xffff) {
    for(cEcmPri *ep=ecmPriList.First(); ep; ep=ecmPriList.Next(ep))
      printf("%s: ecmPriList pri=%d ident=%04x caid=%04x pid=%04x\n",id,ep->pri,ep->sysIdent,ep->ecm->caId,ep->ecm->ecm_pid);
    printf("%s: ecmPri list end\n",id);
    }
#endif
}

// -- cCam ---------------------------------------------------------------

cCam::cCam(const cDevice *dev, int CardNum, int CamDev)
{
  device=(cDevice *)dev;
  cardNum=CardNum;
  camDev=CamDev;

  softCSA=false; numPids=0; source=transponder=-1; logger=0;
  memset(lastCW,0,sizeof(lastCW));
  memset(indexMap,0,sizeof(indexMap));

  if(camDev>=0 && !device->HasDecoder()) { // probably a budget-ci card
    close(camDev); camDev=-1;
    }
  if(camDev<0) {
    if(device->HasDecoder()) LogError("Card %d is a full-featured card but no ca device found!",cardNum);
#ifdef HAVE_SOFTCSA
    softCSA=true;
    LogInfo("Using software decryption on card %d",cardNum);
#endif
    }
}

cCam::~cCam()
{
  handlerList.Clear();
  delete logger;
  if(camDev>0) close(camDev);
}

int cCam::ProvidesCa(const unsigned short *ids)
{
#ifdef DEBUG
  printf("cam %d: check ProvidesCa:",cardNum);
  for(int i=0 ; ids[i]; i++) printf(" %04x",ids[i]);
  printf(" -> ");
#endif
  int n=cSystems::Provides(ids,!softCSA);
  d(printf("n=%d\n",n))
  return n;
}

bool cCam::ConcurrentAllowed(void)
{
  cMutexLock lock(this);
  if(softCSA) return cSoftCAM::ConcurrentAllowed();
  int idx0=FindType(0);
  int idx1=FindType(1);
  int count=0;
  for(cEcmHandler *handler=handlerList.First(); handler; handler=handlerList.Next(handler))
    if(!handler->IsIdle() && ((idx0<0 && idx1<0) || (idx0>=0 && handler->HasPid(pids[idx0])<0) || (idx1>=0 && handler->HasPid(pids[idx1])<0)))
      count++;
  d(printf("cam %d: %d FF recording stream(s) out of %d in use\n",cardNum,count,ScSetup.ConcurrentFF))
  return count<ScSetup.ConcurrentFF;
}

bool cCam::LiveViewActive(void)
{
  cMutexLock lock(this);
  int idx0=FindType(0);
  int idx1=FindType(1);
  int count=0;
  bool live=false;
  for(cEcmHandler *handler=handlerList.First(); handler; handler=handlerList.Next(handler))
    if(!handler->IsIdle()) {
      count++;
      if((idx0>=0 && handler->HasPid(pids[idx0])>=0) || (idx1>=0 && handler->HasPid(pids[idx1])>=0))
        live=true;
      }
  if(count>=ScSetup.ConcurrentFF && live) {
    d(printf("cam %d: active on LiveView and all streams in use\n",cardNum))
    return true;
    }
  return false;
}

void cCam::Tune(int Source, int Transponder)
{
  cMutexLock lock(this);
  if(source!=Source || transponder!=Transponder) {
    while(numPids) {
      d(printf("cam %d: killing leftover pid %04x\n",cardNum,pids[0]))
      DelPid(pids[0]);
      }
    source=Source; transponder=Transponder;
    dp(printf("cam %d: now tuned to source %x transponder %x\n",cardNum,Source,Transponder))
    cSids sids;
    for(cEcmHandler *handler=handlerList.First(); handler; handler=handlerList.Next(handler))
      handler->SetSids(&sids);
    if(logger) logger->Down();
    }
  else dp(printf("cam %d: tune to same source/transponder\n",cardNum))
}

void cCam::AddPid(int pid, int type)
{
#ifdef DEBUG_PIDS
  static const char *typeNames[] = { "AUDIO","VIDEO","PCR","TT","DOLBY","OTHER" };
#define TYPENAME(type) (typeNames[(type)<=5?(type):5])
#endif
  cMutexLock lock(this);
  int idx=FindPid(pid);
  if(idx<0) {
    bool needZero=(!softCSA && ((type<=3) || ScSetup.ConcurrentFF<=1));
    dp(printf("cam %d: adding pid %04x (%s %d zero=%d)\n",cardNum,pid,TYPENAME(type),type,needZero))
    if(type<=3) {
      int x=FindType(type);
      if(x>=0) {
        dp(printf("cam %d: auto-remove pid %04x (%s)\n",cardNum,pids[x],TYPENAME(type)))
        DelPid(pids[x]);
        }
      }
    if(numPids>=MAX_CAM_PIDS) {
      dp(printf("cam %d: pid table overflow\n",cardNum))
      return;
      }
    pids[numPids]=pid; types[numPids]=type; numPids++;

    if(type<=2 || type >=4) {
      cSids sids;
      if(PID2SID(pid,&sids)>0) {
        cEcmHandler *handler=GetHandler(&sids,needZero);
        if(handler) {
          dp(printf("cam %d: found handler for SID %s (%s idle=%d idx=%d)\n",cardNum,*sids.Print(),handler->Id(),handler->IsIdle(),handler->CwIndex()))
          handler->AddPid(pid);
          }
        }
      }
    }
  else {
    dp(printf("cam %d: already has pid %04x (%s %d)\n",cardNum,pid,TYPENAME(types[idx]),types[idx]))
    if(type<2 && types[idx]!=type) {
      dp(printf("cam %d: changing type from %s %d to %s %d for existing pid %04x\n",cardNum,TYPENAME(types[idx]),types[idx],TYPENAME(type),type,pid))
      types[idx]=type;
      }
    }
}

void cCam::DelPid(int pid)
{
  cMutexLock lock(this);
  int idx=FindPid(pid);
  if(idx>=0) {
    dp(printf("cam %d: removing pid %04x\n",cardNum,pid))
    pids[idx]=pids[numPids-1];
    types[idx]=types[numPids-1];
    numPids--;

    cEcmHandler *handler;
    for(handler=handlerList.First(); handler; handler=handlerList.Next(handler))
      if(handler->DelPid(pid)) break;
    }
  else dp(printf("cam %d: does not have pid %04x\n",cardNum,pid))
}

char *cCam::CurrentKeyStr(int num)
{
  cMutexLock lock(this);
  cEcmHandler *handler;
  for(handler=handlerList.First(); handler; handler=handlerList.Next(handler))
    if(--num<0) return handler->CurrentKeyStr();
  return 0;
}

bool cCam::Active(void)
{
  cMutexLock lock(this);
  for(cEcmHandler *handler=handlerList.First(); handler; handler=handlerList.Next(handler))
    if(!handler->IsIdle()) {
      d(printf("cam %d: %s non-idle\n",cardNum,handler->Id()))
      return true;
      }
  return false;
}

void cCam::HouseKeeping(void)
{
  cMutexLock lock(this);
  for(cEcmHandler *handler=handlerList.First(); handler;) {
    cEcmHandler *next=handlerList.Next(handler);
    if(handler->IsRemoveable()) RemHandler(handler);
    handler=next;
    }
  if(handlerList.Count()<1) {
    delete logger; logger=0;
    }
}

void cCam::LogEcmStatus(const cEcmInfo *ecm, bool on)
{
  cMutexLock lock(this);
  if(!logger && on && ScSetup.AutoUpdate) {
    logger=new cLogger(cardNum,softCSA);
#ifdef DEBUG_LOG_STATS
    LogStatsUp();
#endif
    }
  if(logger) logger->EcmStatus(ecm,on);
}

void cCam::SetCWIndex(int pid, int index)
{
  if(index<MAX_CW_IDX) {
    cMutexLock lock(this);
    ca_pid_t ca_pid;
    ca_pid.pid=pid;
    ca_pid.index=index;
    dp(printf("cam %d: descrambling pid %04x on index %x\n",cardNum,pid,index))
    if(!softCSA) { 
      if(ioctl(camDev,CA_SET_PID,&ca_pid)<0)
        if(index>0) {
          LogError("CA_SET_PID failed (%s). Expect a black screen/bad recording. Do you use the patched DVB driver?",strerror(errno));
          LogWarn("Adjusting 'Concurrent streams per FF card' to 1");
          ScSetup.ConcurrentFF=1;
          cSoftCAM::StoreConfig(true);
          }
      }
#ifdef HAVE_SOFTCSA
    else device->SetCaPid(&ca_pid);
#endif
    }
}

void cCam::WriteCW(int index, unsigned char *cw, bool force)
{
  if(index<MAX_CW_IDX) {
    cMutexLock lock(this);
    for(int i=0; i<16; i+=4) cw[i+3]=cw[i]+cw[i+1]+cw[i+2];
    ca_descr_t ca_descr;
    ca_descr.index=index;
    unsigned char *last=lastCW[index];
    if(force || memcmp(&cw[0],&last[0],8)) {
      memcpy(&last[0],&cw[0],8);
      ca_descr.parity=0;
      for(int i=7; i>=0; i--) ca_descr.cw[i]=cw[i];
      if(!softCSA) {
        if(ioctl(camDev,CA_SET_DESCR,&ca_descr)<0)
          LogError("CA_SET_DESCR failed (%s). Expect a black screen.",strerror(errno));
        }
#ifdef HAVE_SOFTCSA
      else device->SetCaDescr(&ca_descr);
#endif
      }

    if(force || memcmp(&cw[8],&last[8],8)) {
      memcpy(&last[8],&cw[8],8);
      ca_descr.parity=1;
      for(int i=7; i>=0; i--) ca_descr.cw[i]=cw[i+8];
      if(!softCSA) {
        if(ioctl(camDev,CA_SET_DESCR,&ca_descr)<0)
          LogError("CA_SET_DESCR failed (%s). Expect a black screen.",strerror(errno));
        }
#ifdef HAVE_SOFTCSA
      else device->SetCaDescr(&ca_descr);
#endif
      }
    }
}

void cCam::Dump(void)
{
#ifdef DEBUG_AV7110
#define CODEBASE (0x2e000404+0x1ce00)
  cMutexLock lock(this);
  if(device->HasDecoder() && lastDump.Elapsed()>20000) {
    lastDump.Set();
    static unsigned int handles=0, hw_handles=0;
    static const unsigned int code[] = {
      0xb5100040,0x4a095a12,0x48094282,0xd00b4b09,0x20000044,
      0x5b1c4294,0xd0033001,0x281cdbf8,0xe001f7fe,0xfd14bc10
      };
    static const unsigned int mask[] = {
      0xffffffff,0xffffffff,0xffffffff,0xffffffff,0xffffffff,
      0xffffffff,0xffffffff,0xffffffff,0xfffff800,0xf800ffff
      };
    if(!handles) {
      handles=1;
      printf("searching handle tables\n");
      for(int i=0; i<0x2000; i+=4) {
        int j;
        for(j=0; j<20; j+=4) {
          int r=av7110_read(camDev,CODEBASE+i+j);
          if((r&mask[j/4])!=(code[j/4]&mask[j/4])) break;
          }
        if(j==20) {
          handles=av7110_read(camDev,CODEBASE+i+44);
          hw_handles=av7110_read(camDev,CODEBASE+i+52);
          printf("found handles=%08x hw_handles=%08x at 0x%08x\n",handles,hw_handles,CODEBASE+i);
          if((handles>>16)!=0x2e08 || (hw_handles>>16)!=0x2e08) {
            printf("seems to be invalid\n");
            }
          break;
          }
        }
      }

    unsigned int hdl=0, hwhdl=0;
    printf("         : 64000080 64000400\n");
    for(int i=0; i<=31; i++) {
      unsigned int off80 =av7110_read(camDev,0x64000080+i*4);
      unsigned int off400=av7110_read(camDev,0x64000400+i*8);
      printf("handle %2d: %08x %08x %s pid=%04x idx=%d",
        i,off80,off400,off80&0x2000?"ACT":"---",off80&0x1fff,(off400&0x1e)>>1);
      if(handles>1 && i<=27) {
        if((i&1)==0) {
          hdl=av7110_read(camDev,handles+i*2);
          hwhdl=av7110_read(camDev,hw_handles+i*2);
          }
        unsigned int s=((~i)&1)<<4;
        printf(" | %02d hdl=%04x hwfilt=%04x",i,(hdl>>s)&0xffff,(hwhdl>>s)&0xffff);
        }
      printf("\n");
      }
    }
#endif
}

int cCam::FindPid(int pid)
{
  for(int idx=0; idx<numPids; idx++)
    if(pids[idx]==pid) return idx;
  return -1;
}

int cCam::FindType(int type)
{
  for(int idx=0; idx<numPids; idx++)
    if(types[idx]==type) return idx;
  return -1;
}

int cCam::GetFreeIndex(void)
{
  for(int idx=0; idx<MAX_CW_IDX; idx++)
    if(!indexMap[idx]) return idx;
  return -1;
}

cEcmHandler *cCam::GetHandler(cSids *sids, bool needZero)
{
  cEcmHandler *zerohandler=0, *sidhandler=0, *idlehandler=0;
  for(cEcmHandler *handler=handlerList.First(); handler; handler=handlerList.Next(handler)) {
    if(sids->Compare(handler->Sids(),true))
      sidhandler=handler;
    if(handler->CwIndex()==0)
      zerohandler=handler;
    if(handler->IsIdle() && (!idlehandler || (!(needZero ^ (idlehandler->CwIndex()!=0)) && (needZero ^ (handler->CwIndex()!=0))) ))
      idlehandler=handler;
    }
#ifdef DEBUG_PIDS
  printf("cam %d: SID=%s zero=%d | ",cardNum,*sids->Print(),needZero);
  if(sidhandler) printf(" sid=%d/%s/%d",sidhandler->CwIndex(),*sidhandler->Sids()->Print(),sidhandler->IsIdle());
  else printf(" sid=-/-/-");
  if(zerohandler) printf(" zero=%d/%s/%d",zerohandler->CwIndex(),*zerohandler->Sids()->Print(),zerohandler->IsIdle());
  else printf(" zero=-/-/-");
  if(idlehandler) printf(" idle=%d/%s/%d",idlehandler->CwIndex(),*idlehandler->Sids()->Print(),idlehandler->IsIdle());
  else printf(" idle=-/-/-");
  printf("\n");
#endif

  if(sidhandler) {
    if(needZero && sidhandler->CwIndex()!=0) {
      if(!sidhandler->IsIdle())
        d(printf("cam %d: shifting cwindex on non-idle handler.\n",cardNum))
      if(zerohandler) {
        if(!zerohandler->IsIdle())
          d(printf("cam %d: shifting non-idle zero handler. This shouldn't happen!\n",cardNum))
        zerohandler->ShiftCwIndex(sidhandler->CwIndex());
        sidhandler->ShiftCwIndex(0);
        }
      else if(indexMap[0]==0) {
        indexMap[0]=1;
        indexMap[sidhandler->CwIndex()]=0;
        sidhandler->ShiftCwIndex(0);
        }
      else d(printf("cam %d: zero index not free.\n",cardNum))
      }

    if(!needZero && sidhandler->CwIndex()==0) {
      if(!sidhandler->IsIdle())
        d(printf("cam %d: shifting cwindex on non-idle handler.\n",cardNum))
      int idx=GetFreeIndex();
      if(idx>=0) {
        indexMap[idx]=1;
        sidhandler->ShiftCwIndex(idx);
        indexMap[0]=0;
        }
      else d(printf("cam %d: no free cwindex. Can't free zero index.\n",cardNum))
      }

    return sidhandler;
    }

  if(needZero && zerohandler) {
    if(!zerohandler->IsIdle())
      d(printf("cam %d: changing SID on non-idle zero handler. This shouldn't happen!\n",cardNum))
    zerohandler->SetSids(sids);
    return zerohandler;
    }
  
  if(idlehandler) {
    idlehandler->SetSids(sids);
    if(needZero && idlehandler->CwIndex()!=0) {
      if(indexMap[0]==0) {
        indexMap[0]=1;
        indexMap[idlehandler->CwIndex()]=0;
        idlehandler->ShiftCwIndex(0);
        }
      else d(printf("cam %d: zero index not free. (2)\n",cardNum))
      }
    if(!needZero && idlehandler->CwIndex()==0) {
      int idx=GetFreeIndex();
      if(idx>=0) {
        indexMap[idx]=1;
        idlehandler->ShiftCwIndex(idx);
        indexMap[0]=0;
        }
      else d(printf("cam %d: no free cwindex. Can't free zero index. (2)\n",cardNum))
      }
    if((needZero ^ (idlehandler->CwIndex()==0)))
      d(printf("cam %d: idlehandler index doesn't match needZero\n",cardNum))
    return idlehandler;
    }

  int idx=GetFreeIndex();
  if(!needZero && idx==0) {
    indexMap[0]=1;
    idx=GetFreeIndex();
    indexMap[0]=0;
    if(idx<0) {
      idx=0;
      d(printf("cam %d: can't respect !needZero for new handler\n",cardNum))
      }
    }
  if(idx<0) {
    d(printf("cam %d: no free cwindex\n",cardNum))
    return 0;
    }
  indexMap[idx]=1;
  idlehandler=new cEcmHandler(this,cardNum,sids,idx);
  handlerList.Add(idlehandler);
  return idlehandler;
}

void cCam::RemHandler(cEcmHandler *handler)
{
  int idx=handler->CwIndex();
  dp(printf("cam %d: removing %s on cw index %d\n",cardNum,handler->Id(),idx))
  handlerList.Del(handler);
  indexMap[idx]=0;
}

int cCam::PID2SID(int pid, cSids *sids)
{
  Channels.Lock(false);
  sids->Clear();
  for(cChannel *channel=Channels.First(); channel; channel=Channels.Next(channel)) {
    if(!channel->GroupSep() &&
        channel->Ca()>=CA_ENCRYPTED_MIN &&
        channel->Source()==source &&
        ISTRANSPONDER(channel->Transponder(),transponder)) {
      bool done=false;
      if(channel->Vpid()==pid) done=true;
      if(!done) {
        for(int i=0,p ; (p=channel->Apid(i)); i++)
          if(p==pid) { done=true; break; }
        }
      if(!done) {
        for(int i=0,p ; (p=channel->Dpid(i)); i++)
          if(p==pid) { done=true; break; }
        }
      if(!done) {
        for(int i=0,p ; (p=channel->Spid(i)); i++)
          if(p==pid) { done=true; break; }
        }
      if(done && device->CiHandler()) {
        unsigned short ids[MAXCAIDS+1];
        for(int i=0 ; i<=MAXCAIDS; i++) if((ids[i]=channel->Ca(i))==0) break;
        if(cSystems::Provides(ids,!softCSA)<=0) done=false;
        }
      if(done) {
        sids->AddSid(channel->Sid());
        dp(printf("cam %d: mapping pid %04x to SID %d on channel %s\n",cardNum,pid,channel->Sid(),channel->Name()))
        }
      }
    }
  Channels.Unlock();
  return sids->Count();
}
